from .unpacking import (
    doublestarred,
    doublestarredpart,
    starred,
    starredpart,
    unpacking,
    unpackingpart,
)

star = starred
dblstar = doublestarred
unpk = unpacking

starpr = starredpart
dblstarpr = doublestarredpart
unpkpr = unpackingpart



from collections.abc import Callable, Iterable, Mapping
from functools import partial as prt
from inspect import signature
from itertools import islice
from operator import contains

from cytoolz import keyfilter  # from toolz import keyfilter  # type: ignore[import]
from plum import dispatch


@dispatch
def apply_packed[T](fnct: Callable[..., T], itrb: Iterable) -> T:  # type: ignore[return]
    return fnct(*itrb)


@dispatch
def apply_packed[T](fnct: Callable[..., T], assctbl: Mapping) -> T:
    return fnct(**assctbl)


@dispatch
def apply_packed_part[T](fnct: Callable[..., T], itrb: Iterable) -> T:  # type: ignore[return]
    return fnct(*islice(itrb, len(signature(fnct).parameters)))


@dispatch
def apply_packed_part[T](fnct: Callable[..., T], assctbl: Mapping) -> T:
    return fnct(**keyfilter(prt(contains, signature(fnct).parameters.keys()), assctbl))


class starred[T]:
    def __init__(self, fnct: Callable[..., T]):
        self.fnct = fnct

    def __call__(self, itrb: Iterable) -> T:
        return self.fnct(*itrb)


class doublestarred[T]:
    def __init__(self, fnct: Callable[..., T]):
        self.fnct = fnct

    def __call__(self, assctbl: Mapping) -> T:
        return self.fnct(**assctbl)


class unpacking[T]:
    def __init__(self, fnct: Callable[..., T]):
        self.fnct = fnct

    def __call__(self, x: Iterable | Mapping) -> T:
        return apply_packed(self.fnct, x)  # type: ignore[return]


class starredpart[T]:
    def __init__(self, fnct: Callable[..., T]):
        self.fnct = fnct

    def __call__(self, itrb: Iterable) -> T:
        return self.fnct(*islice(itrb, len(signature(self.fnct).parameters)))


class doublestarredpart[T]:
    def __init__(self, fnct: Callable[..., T]):
        self.fnct = fnct

    def __call__(self, assctbl: Mapping) -> T:
        return self.fnct(
            **keyfilter(prt(contains, signature(self.fnct).parameters.keys()), assctbl)
        )


class unpackingpart[T]:
    def __init__(self, fnct: Callable[..., T]):
        self.fnct = fnct

    def __call__(self, x: Iterable | Mapping) -> T:
        return apply_packed_part(self.fnct, x)  # type: ignore[return]



import unittest
from concurrent.futures import ProcessPoolExecutor

from src.unpacking import (
    doublestarred,
    doublestarredpart,
    starred,
    starredpart,
    unpacking,
    unpackingpart,
)


def add(x, y):
    return x + y


class TestUnpacking(unittest.TestCase):
    def test_basic(self):

        args = [1, 2]
        kwargs = {"x": 1, "y": 2}
        self.assertEqual(starred(add)(args), 3)
        self.assertEqual(doublestarred(add)(kwargs), 3)
        self.assertEqual(unpacking(add)(args), 3)
        self.assertEqual(unpacking(add)(kwargs), 3)

    def test_multiprocessing(self):

        args_list = [[1, 2], [3, 4]]
        kwargs_list = [{"x": 1, "y": 2}, {"x": 3, "y": 4}]
        with ProcessPoolExecutor(2) as executor:
            result_args = tuple(executor.map(unpacking(add), args_list))
            self.assertEqual(result_args, (3, 7))
            result_kwargs = tuple(executor.map(unpacking(add), kwargs_list))
            self.assertEqual(result_kwargs, (3, 7))

    def test_excess_args(self):

        args_excess = [1, 2, 3]
        kwargs_excess = {"x": 1, "y": 2, "z": 3}
        self.assertEqual(starredpart(add)(args_excess), 3)
        self.assertEqual(doublestarredpart(add)(kwargs_excess), 3)
        args_excess_m = [[1, 2, 3], [3, 4, 5]]
        kwargs_excess_m = [{"x": 1, "y": 2, "z": 3}, {"x": 3, "y": 4, "z": 5}]
        with ProcessPoolExecutor(2) as executor:
            result_args = tuple(executor.map(unpackingpart(add), args_excess_m))
            self.assertEqual(result_args, (3, 7))
            result_kwargs = tuple(executor.map(unpackingpart(add), kwargs_excess_m))
            self.assertEqual(result_kwargs, (3, 7))


if __name__ == "__main__":
    unittest.main()
